#!/usr/bin/env python3
#
#  Copyright (c) 2020 Two Sigma Open Source, LLC
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.
#
"""Module implementing the Mesos file access REST API to serve Cook job logs. """

import logging
import os
import signal
import sys
from operator import itemgetter
from pathlib import Path
from stat import *

import gunicorn.app.base
import gunicorn.arbiter
from flask import Flask, jsonify, request, send_file

from cook.sidecar import util
from cook.sidecar.version import VERSION

app = Flask(__name__)
sandbox_directory = None
max_read_length = int(os.environ.get('COOK_FILE_SERVER_MAX_READ_LENGTH', '25000000'))


def start_file_server(started_event, args):
    try:
        logging.info(f'Starting cook.sidecar {VERSION} file server')
        port, workers, threads = (args + [None] * 3)[0:3]
        if port is None:
            logging.error('Must provide file server port')
            sys.exit(1)
        cook_workdir = os.environ.get('COOK_WORKDIR')
        if not cook_workdir:
            logging.error('COOK_WORKDIR environment variable must be set')
            sys.exit(1)
        FileServerApplication(cook_workdir, started_event, {
            'bind': f'0.0.0.0:{port}',
            'threads': 2 if threads is None else threads,
            'workers': 4 if workers is None else workers,
        }).run()
        return 0

    except Exception as e:
        logging.exception(f'exception when running file server with {args}')
        return 1


class FileServerArbiter(gunicorn.arbiter.Arbiter):
    '''
    Custom Gunicorn Arbiter object,
    with overridden logic for re-installing existing signal handlers,
    and an event to indicate when the server is up and running.
    '''

    def __init__(self, app, started_event):
        self.started_event = started_event
        # Since gunicorn installs its own signal handler routine for basically all signals,
        # we need to save any existing signal handlers that we want preserved and then
        # inject them back into gunicorn's hanlder (injected via the `signal` method below).
        self.user_signal_handlers = {}
        for sig in range(1, signal.NSIG):
            user_handler = signal.getsignal(sig)
            if callable(user_handler):
                logging.info(f'Saving user handler for signal {sig}')
                self.user_signal_handlers[sig] = user_handler
        super().__init__(app)

    def start(self):
        super().start()
        # Gunicorn invokes the `when_ready` hook at the end of the Arbiter's `start` method.
        # https://docs.gunicorn.org/en/stable/settings.html#when-ready
        # However, since we already needed a custom Arbiter for signal handling logic,
        # providing that custom hook was much more complex than signaling
        # here that the file server has started than additionally providing a custom Config object.
        logging.info(f'Sidecar file server is ready')
        self.started_event.set()

    def signal(self, sig, frame):
        '''Generic signal handler for all signals, installed by gunicorn.'''
        # Invoke the user's handler for the signal (if present).
        # See comment in `__init__` above for more details.
        user_handler = self.user_signal_handlers.get(sig)
        if user_handler is not None:
            logging.info(f'Entering user handler for signal {sig}')
            user_handler(sig, frame)
            logging.info(f'Exiting user handler for signal {sig}')
        # Enter gunicorn's handler for the signal.
        super().signal(sig, frame)


class FileServerApplication(gunicorn.app.base.BaseApplication):

    def __init__(self, cook_workdir, started_event, options=None):
        self.options = options or {}
        self.application = app
        self.started_event = started_event
        global sandbox_directory
        sandbox_directory = cook_workdir
        super(FileServerApplication, self).__init__()

    def load_config(self):
        for key, value in self.options.items():
            if key in self.cfg.settings and value is not None:
                self.cfg.set(key.lower(), value)

    def load(self):
        return self.application

    def run(self):
        try:
            FileServerArbiter(self, self.started_event).run()
        except RuntimeError as e:
            logging.exception('Error while running cook.sidecar file server')


def path_is_valid(path):
    if not os.path.exists(path):
        return False
    normalized_path = os.path.normpath(path)
    return normalized_path.startswith(sandbox_directory)


@app.route('/files/download')
@app.route('/files/download.json')
def download():
    path = request.args.get('path')
    if path is None:
        return "Expecting 'path=value' in query.\n", 400
    if not path_is_valid(path):
        return "", 404
    if os.path.isdir(path):
        return "Cannot download a directory.\n", 400
    return send_file(path, as_attachment=True)


@app.route('/files/read')
@app.route('/files/read.json')
def read():
    path = request.args.get('path')
    offset_param = request.args.get('offset', -1)
    length_param = request.args.get('length', -1)
    if path is None:
        return "Expecting 'path=value' in query.\n", 400
    try:
        offset = int(offset_param)
    except ValueError as _:
        return f"Failed to parse offset: Failed to convert '{offset_param}' to number.\n", 400
    if offset < -1:
        return f"Negative offset provided: {offset_param}.\n", 400
    try:
        length = int(length_param)
    except ValueError as _:
        return f"Failed to parse length: Failed to convert '{length_param}' to number.\n", 400
    if length < -1:
        return f"Negative length provided: {length_param}.\n", 400
    if not path_is_valid(path):
        return "", 404
    if os.path.isdir(path):
        return "Cannot read a directory.\n", 400
    if offset == -1:
        return jsonify({
            "data": "",
            "offset": os.path.getsize(path),
        })
    f = open(path)
    f.seek(offset)
    length = max_read_length if length == -1 else length
    if length > max_read_length:
        return f"Requested length for file read, {length} is greater than max allowed length, {max_read_length}", 400
    data = f.read(length)
    f.close()
    return jsonify({
        "data": data,
        "offset": offset,
    })


def make_permission_string(permission_bits):
    return ''.join(["rwxrwxrwx"[i] if (permission_bits & (1 << (8 - i)) != 0) else "-" for i in range(0, 9)])


@app.route('/files/browse')
@app.route('/files/browse.json')
def browse():
    path = request.args.get('path')
    if path is None:
        return "Expecting 'path=value' in query.\n", 400
    if not path_is_valid(path):
        return "", 404
    if not os.path.isdir(path):
        return jsonify([])
    retval = [
        {
            "gid": path_obj.group(),
            "mode": ('d' if S_ISDIR(st.st_mode) else '-') + make_permission_string(S_IMODE(st.st_mode) % 512),
            "mtime": int(st.st_mtime),
            "nlink": st.st_nlink,
            "path": path,
            "size": st.st_size,
            "uid": path_obj.owner(),
        }
        for st, path_obj, path in [(os.stat(path), Path(path), path)
                                   for path in [os.path.join(path, f)
                                                for f in os.listdir(path)]]
    ]
    return jsonify(sorted(retval, key=itemgetter("path")))


# This endpoint is not part of the Mesos API. It is used by the kubernetes readiness probe on the sidecar container.
# Cook will see that the sidecar is ready to serve files and will set the output_url. If we expose the output_url
# before the server is ready, then someone might use it and get an error.
@app.route('/readiness-probe')
def readiness_probe():
    return ""


def main():
    util.init_logging()
    if len(sys.argv) == 2 and sys.argv[1] == "--version":
        print(VERSION)
    else:
        start_file_server(sys.argv[1:])


if __name__ == '__main__':
    main()



#
#  Copyright (c) 2020 Two Sigma Open Source, LLC
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.
#
"""Configuration loading for Cook's sidecar progress reporter."""

import logging
import os


class ProgressReporterConfig(object):
    """This class is responsible for storing the progress reporter config."""

    def __init__(self,
                 callback_url,
                 max_bytes_read_per_line,
                 max_message_length,
                 max_post_attempts,
                 max_post_time_secs,
                 progress_output_env_variable,
                 progress_output_name,
                 progress_regex_string,
                 progress_sample_interval_ms,
                 sandbox_directory):
        self.callback_url = callback_url
        self.max_bytes_read_per_line = max_bytes_read_per_line
        self.max_message_length = max_message_length
        self.max_post_attempts = max_post_attempts
        self.max_post_time_secs = max_post_attempts
        self.progress_output_env_variable = progress_output_env_variable
        self.progress_output_name = progress_output_name
        self.progress_regex_string = progress_regex_string
        self.progress_sample_interval_ms = progress_sample_interval_ms
        self.sandbox_directory = sandbox_directory

    def sandbox_file(self, file):
        return os.path.join(self.sandbox_directory, file)

    def stderr_file(self):
        return self.sandbox_file('stderr')

    def stdout_file(self):
        return self.sandbox_file('stdout')


def initialize_config(environment):
    """Initializes the config using the environment.
    Populates the default values for missing environment variables.
    """
    instance_id = environment.get('COOK_INSTANCE_UUID')
    if instance_id is None:
        raise Exception('Task unknown! COOK_INSTANCE_UUID not set in environment.')
    job_uuid = environment.get('COOK_JOB_UUID')
    if job_uuid is None:
        raise Exception('Job unknown! job_uuid not set in environment.')

    cook_scheduler_rest_url = environment.get('COOK_SCHEDULER_REST_URL')
    if cook_scheduler_rest_url is None:
        raise Exception('REST URL unknown! COOK_SCHEDULER_REST_URL not set in environment.')

    callback_url = f'{cook_scheduler_rest_url}/progress/{instance_id}'

    sandbox_directory = environment.get('COOK_WORKDIR', '')

    default_progress_output_key = 'EXECUTOR_PROGRESS_OUTPUT_FILE'
    progress_output_env_key = 'EXECUTOR_PROGRESS_OUTPUT_FILE_ENV'
    progress_output_env_variable = environment.get(progress_output_env_key, default_progress_output_key)
    logging.info(f'Progress location environment variable is {progress_output_env_variable}')
    if progress_output_env_variable not in environment:
        logging.info(f'No entry found for {progress_output_env_variable} in the environment')

    default_progress_output_name = environment.get(progress_output_env_variable, f'{job_uuid}.progress')
    if sandbox_directory:
        default_progress_output_file = os.path.join(sandbox_directory, default_progress_output_name)
    else:
        default_progress_output_file = default_progress_output_name

    max_bytes_read_per_line = max(int(environment.get('EXECUTOR_MAX_BYTES_READ_PER_LINE', 4 * 1024)), 128)
    max_message_length = max(int(environment.get('EXECUTOR_MAX_MESSAGE_LENGTH', 512)), 64)
    max_post_attempts = max(int(environment.get('PROGRESS_MAX_POST_ATTEMPTS', 5)), 1)
    max_post_time_secs = float(environment.get('PROGRESS_MAX_POST_TIME_SECS', 10.0))
    progress_output_name = environment.get(progress_output_env_variable, default_progress_output_file)
    progress_regex_string = environment.get('PROGRESS_REGEX_STRING', r'progress: ([0-9]*\.?[0-9]+), (.*)')
    progress_sample_interval_ms = max(int(environment.get('PROGRESS_SAMPLE_INTERVAL_MS', 1000)), 100)
    sandbox_directory = environment.get('COOK_WORKDIR', '')

    if sandbox_directory and not progress_output_name.startswith('/'):
        progress_output_name = os.path.join(sandbox_directory, progress_output_name)

    logging.info(f'Job instance UUID is {instance_id}')
    logging.info(f'Progress update callback url is {callback_url}')
    logging.info(f'Max bytes read per line is {max_bytes_read_per_line}')
    logging.info(f'Progress message length is limited to {max_message_length}')
    logging.info(f'Progress post attempts limit is {max_post_attempts}')
    logging.info(f'Post request timeout {max_post_time_secs}')
    logging.info(f'Progress output file is {progress_output_name}')
    logging.info(f'Progress regex is {progress_regex_string}')
    logging.info(f'Progress sample interval is {progress_sample_interval_ms}')
    logging.info(f'Sandbox location is {sandbox_directory}')

    return ProgressReporterConfig(callback_url=callback_url,
                                  max_bytes_read_per_line=max_bytes_read_per_line,
                                  max_message_length=max_message_length,
                                  max_post_attempts=max_post_attempts,
                                  max_post_time_secs=max_post_time_secs,
                                  progress_output_env_variable=progress_output_env_variable,
                                  progress_output_name=progress_output_name,
                                  progress_regex_string=progress_regex_string,
                                  progress_sample_interval_ms=progress_sample_interval_ms,
                                  sandbox_directory=sandbox_directory)



#!/usr/bin/env python3
#
#  Copyright (c) 2020 Two Sigma Open Source, LLC
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.
#
"""Entrypoint for Cook's sidecar progress reporter."""

import faulthandler
import logging
import os
import requests
import signal
import sys

import cook.sidecar.config as csc
import cook.sidecar.tracker as cst
from cook.sidecar import util
from cook.sidecar.version import VERSION


def start_progress_trackers():
    try:
        logging.info(f'Starting cook.sidecar {VERSION} progress reporter')
        config = csc.initialize_config(os.environ)

        default_url = config.callback_url
        current_url = default_url

        def send_progress_message(message):
            nonlocal current_url
            try:
                for i in range(config.max_post_attempts):
                    response = requests.post(current_url, allow_redirects=False, timeout=config.max_post_time_secs, json=message)
                    if 200 <= response.status_code <= 299:
                        return True
                    elif response.is_redirect and response.status_code == 307:
                        current_url = response.headers['location']
                        logging.info(f'Redirected! Changed progress update callback url to: {current_url}')
                    else:
                        logging.warning(f'Unexpected progress update response ({response.status_code}): {response.content}')
                        break
                else:
                    logging.warning(f'Reached max redirect retries ({config.max_post_redirect_follow})')
            except Exception:
                logging.exception(f'Error raised while posting progress update to {current_url}')
            current_url = default_url
            logging.info(f'Failed to post progress update. Reset progress update callback url: {current_url}')
            return False

        max_message_length = config.max_message_length
        sample_interval_ms = config.progress_sample_interval_ms
        sequence_counter = cst.ProgressSequenceCounter()
        progress_updater = cst.ProgressUpdater(max_message_length, sample_interval_ms, send_progress_message)

        def launch_progress_tracker(progress_location, location_tag):
            progress_file_path = os.path.abspath(progress_location)
            logging.info(f'Location {progress_location} (absolute path={progress_file_path}) tagged as [tag={location_tag}]')
            progress_tracker = cst.ProgressTracker(config, sequence_counter, progress_updater, progress_location, location_tag)
            progress_tracker.start()
            return progress_tracker

        progress_locations = {config.progress_output_name: 'progress',
                              config.stderr_file(): 'stderr',
                              config.stdout_file(): 'stdout'}
        logging.info(f'Progress will be tracked from {len(progress_locations)} locations')
        progress_trackers = [launch_progress_tracker(file, name) for file, name in progress_locations.items()]

        def set_terminate_handler(handler):
            signal.signal(signal.SIGINT, handler)
            signal.signal(signal.SIGTERM, handler)

        def exit_on_interrupt(interrupt_code, _):
            sys.exit(f'Progress Reporter killed with code {interrupt_code}')

        def handle_interrupt(interrupt_code, _):
            logging.info(f'Progress Reporter interrupted with code {interrupt_code}')
            # allow a second signal to kill the process immediately (no blocking)
            set_terminate_handler(exit_on_interrupt)
            # force send the latest progress state if available, and stop the tracker
            for progress_tracker in progress_trackers:
                progress_tracker.stop()
            for progress_tracker in progress_trackers:
                progress_tracker.wait()

        def dump_traceback(signal, frame):
            faulthandler.dump_traceback()

        signal.signal(signal.SIGUSR1, dump_traceback)
        set_terminate_handler(handle_interrupt)

        return progress_trackers

    except Exception:
        logging.exception('Error starting Progress Reporter')
        return None


def await_progress_trackers(progress_trackers):
        if progress_trackers is None:
            sys.exit('Failed to start progress trackers')
        # wait for all background threads to exit
        # (but this process will probably be killed first instead)
        for progress_tracker in progress_trackers:
            progress_tracker.wait()


def main():
    util.init_logging()
    if len(sys.argv) == 2 and sys.argv[1] == "--version":
        print(VERSION)
    else:
        progress_trackers = start_progress_trackers()
        await_progress_trackers(progress_trackers)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
#  Copyright (c) 2020 Two Sigma Open Source, LLC
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.
#

import argparse
import logging
import sys
import threading

from cook.sidecar import exit_sentinel, file_server, progress, util
from cook.sidecar.version import VERSION


def main(args=None):
    util.init_logging()

    parser = argparse.ArgumentParser(description='Cook Sidecar')
    parser.add_argument('--exit-sentinel-file-path', metavar='PATH', help='file path which signals this process to exit when it appears')
    parser.add_argument('--file-server-port', type=int, metavar='PORT', help='file server port number')
    parser.add_argument('--file-server-threads', type=int, default=2, metavar='THREADS', help='file server threads-per-worker count')
    parser.add_argument('--file-server-workers', type=int, default=2, metavar='WORKERS', help='file server worker process count')
    parser.add_argument('--no-file-server', action='store_true', help='disable sandbox file server')
    parser.add_argument('--no-progress-reporter', action='store_true', help='disable progress reporter')
    parser.add_argument('--version', action='version', version=f'Cook Sidecar {VERSION}')
    options = parser.parse_args(args or sys.argv[1:])
    logging.info(f'OPTIONS = {options}')

    logging.info(f'Starting cook.sidecar {VERSION}')

    all_started_event = threading.Event()
    exit_code = 0

    # Start progress reporter workers (non-blocking)
    if not options.no_progress_reporter:
        progress_trackers = progress.start_progress_trackers()

    # Start exit sentinel file watcher thread
    if options.exit_sentinel_file_path:
        exit_sentinel.watch_for_file(options.exit_sentinel_file_path, all_started_event)

    # Start Flask file server (blocking)
    if not options.no_file_server:
        file_server_args = [options.file_server_port, options.file_server_workers, options.file_server_threads]
        exit_code = file_server.start_file_server(all_started_event, file_server_args)
    # Wait for progress reporter threads (blocking)
    elif not options.no_progress_reporter:
        all_started_event.set()
        progress.await_progress_trackers(progress_trackers)

    # Propagate file server's exit code
    sys.exit(exit_code)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
#  Copyright (c) 2020 Two Sigma Open Source, LLC
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.
#
"""Cook sidecar exit sentinel file watcher thread logic."""

import logging
import os
import signal
import threading
import time

def watch_for_file(sentinel_file_path, started_event):
    def daemon_routine():
        # wait for other components to finish starting
        logging.info(f'Waiting for all components to start...')
        started_event.wait()
        # wait for sentinel file to appear
        logging.info(f'Watching for sentinel file: {sentinel_file_path}')
        while not os.path.exists(sentinel_file_path):
            time.sleep(0.1)
        # trigger this process's termination handler
        logging.info(f'Sidecar termination triggered by sentinel file: {sentinel_file_path}')
        os.kill(os.getpid(), signal.SIGTERM)
    threading.Thread(target=daemon_routine, args=(), daemon=True).start()



#
#  Copyright (c) 2020 Two Sigma Open Source, LLC
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.
#
"""Implementation for Cook's sidecar progress reporter."""

import logging
import os
import re
import time
from threading import Event, Lock, Thread


class ProgressSequenceCounter:
    """Utility class that supports atomically incrementing the sequence value."""
    def __init__(self, initial=0):
        self.lock = Lock()
        self.value = initial

    def increment_and_get(self):
        """Atomically increments by one the current value and returns the new value."""
        with self.lock:
            self.value += 1
            return self.value


class ProgressUpdater(object):
    """This class is responsible for sending progress updates to the scheduler.
    It throttles the rate at which progress updates are sent.
    """

    def __init__(self, max_message_length, poll_interval_ms, send_progress_message_fn):
        """
        max_message_length: int
            The allowed max message length after encoding.
        poll_interval_ms: int
            The interval after which to send a subsequent progress update.
        send_progress_message_fn: function(message)
            The helper function used to send the progress message.
        """
        self.max_message_length = max_message_length
        self.poll_interval_ms = poll_interval_ms
        self.last_reported_time = None
        self.last_progress_data_sent = None
        self.send_progress_message = send_progress_message_fn
        self.lock = Lock()

    def has_enough_time_elapsed_since_last_update(self):
        """Returns true if enough time (based on poll_interval_ms) has elapsed since
        the last progress update (available in last_reported_time).
        """
        if self.last_reported_time is None:
            return True
        else:
            current_time = time.time()
            time_diff_ms = (current_time - self.last_reported_time) * 1000
            return time_diff_ms >= self.poll_interval_ms

    def is_increasing_sequence(self, progress_data):
        """Checks if the sequence number in progress_data is larger than the previously published progress.

        Parameters
        ----------
        progress_data: dictionary
            The progress data to send.

        Returns
        -------
        True if the sequence number in progress_data is larger than the previously published progress, False otherwise
        """
        last_progress_data = self.last_progress_data_sent
        last_progress_sequence = last_progress_data['progress-sequence'] if last_progress_data else -1
        return progress_data['progress-sequence'] > last_progress_sequence

    def send_progress_update(self, progress_data, force_send=False):
        """Sends a progress update if enough time has elapsed since the last progress update.
        The force_send flag can be used to ignore the check for enough time having elapsed.
        Using this method is thread-safe.

        Parameters
        ----------
        progress_data: dictionary
            The progress data to send.
        force_send: boolean, optional
            Defaults to false.

        Returns
        -------
        Nothing
        """
        with self.lock:
            # ensure we do not send outdated progress data due to parallel repeated calls to this method
            if progress_data is None or not self.is_increasing_sequence(progress_data):
                logging.info(f'Skipping invalid/outdated progress data {progress_data}')
            elif not force_send and not self.has_enough_time_elapsed_since_last_update():
                logging.debug('Not sending progress data as enough time has not elapsed since last update')
            else:
                logging.info(f'Sending progress message {progress_data}')
                message_dict = dict(progress_data)

                raw_progress_message = progress_data['progress-message']
                try:
                    progress_str = raw_progress_message.decode('ascii').strip()
                except UnicodeDecodeError:
                    logging.info('Unable to decode progress message in ascii, using empty string instead')
                    progress_str = ''

                if len(progress_str) <= self.max_message_length:
                    message_dict['progress-message'] = progress_str
                else:
                    allowed_progress_message_length = max(self.max_message_length - 3, 0)
                    new_progress_str = progress_str[:allowed_progress_message_length].strip() + '...'
                    logging.info(f'Progress message trimmed to {new_progress_str}')
                    message_dict['progress-message'] = new_progress_str

                send_success = self.send_progress_message(message_dict)
                if send_success:
                    self.last_progress_data_sent = progress_data
                    self.last_reported_time = time.time()
                else:
                    logging.info(f'Unable to send progress message {message_dict}')


class ProgressWatcher(object):
    """This class tails the output from the target file listening for progress messages.
    The retrieve_progress_states generates all progress messages iteratively.
    """

    def __init__(self, output_name, location_tag, sequence_counter,
                 max_bytes_read_per_line, progress_regex_string, stop_event):
        """The ProgressWatcher constructor.

        Parameters
        ----------
        progress_regex_string: string
            The progress regex to match against, it must return one or two capture groups.
            The first capture group represents the progress percentage.
            The second capture group, if present, represents the progress message.
        """
        self.target_file = output_name
        self.location_tag = location_tag
        self.sequence_counter = sequence_counter
        self.max_bytes_read_per_line = max_bytes_read_per_line
        self.progress_regex_string = progress_regex_string
        self.progress_regex_pattern = re.compile(progress_regex_string.encode())
        self.progress = None
        self.stop_event = stop_event

    def stopped(self):
        """Check if this progress tracker has been stopped."""
        return self.stop_event.is_set()

    def current_progress(self):
        """Returns the current progress dictionary."""
        return self.progress

    def tail(self, sleep_time_ms):
        """This method incrementally generates lines from a file by waiting for new content from a file.
        It behaves like the 'tail -f' shell command.

        Parameters
        ----------
        sleep_time_ms: int
            The unit of time in ms to repetitively sleep when the file has not been created or no new
            content is available in the file being tailed.

        Returns
        -------
        an incrementally generated list of lines in the file being tailed.
        """
        try:
            sleep_param = sleep_time_ms / 1000

            if os.path.exists(self.target_file) and not os.path.isfile(self.target_file):
                logging.info(f'Skipping progress monitoring on {self.target_file} as it is not a file')
                return

            if not os.path.isfile(self.target_file):
                logging.debug(f'Awaiting creation of file {self.target_file} [tag={self.location_tag}]')

            while not os.path.isfile(self.target_file):
                if self.stopped():
                    return
                time.sleep(sleep_param)

            if not os.path.isfile(self.target_file):
                logging.info(f'Progress output file has not been created [tag={self.location_tag}]')
                return

            logging.info(f'File has been created, reading contents [tag={self.location_tag}]')
            linesep_bytes = os.linesep.encode()
            fragment_index = 0
            line_index = 0

            post_stop_bytes_to_read = 2000000  # read max of 2 MB after stop signal comes in

            with open(self.target_file, 'rb') as target_file_obj:
                while True:
                    line = target_file_obj.readline(self.max_bytes_read_per_line)

                    if self.stopped():
                        # exit generator once input or post_stop_bytes_to_read is exhausted
                        if not line or post_stop_bytes_to_read <= 0:
                            return
                        # limit how much data to process after stop signal received
                        else:
                            post_stop_bytes_to_read -= len(line)

                    # no new data available, sleep before trying again
                    if not line:
                        time.sleep(sleep_param)
                        continue

                    fragment_index += 1
                    if line.endswith(linesep_bytes):
                        line_index += 1
                    yield line

        except Exception as exception:
            logging.exception(f'Error while tailing {self.target_file} [tag={self.location_tag}]')
            raise exception

    def match_progress_update(self, input_data):
        """Returns the progress tuple when the input string matches the provided regex.

        Parameters
        ----------
        input_data: bytes
            The input data.

        Returns
        -------
        the tuple (percent, message) if the string matches the provided regex,
                 else return None.
        """
        matches = self.progress_regex_pattern.findall(input_data)
        return matches[0] if len(matches) >= 1 else None

    def __update_progress(self, progress_report):
        """Updates the progress field with the data from progress_report if it is valid."""
        if isinstance(progress_report, tuple) and len(progress_report) == 2:
            percent_data, message_data = progress_report
        elif isinstance(progress_report, tuple) and len(progress_report) == 1:
            percent_data, message_data = progress_report[0], b''
        else:
            percent_data, message_data = progress_report, b''

        percent_float = float(percent_data.decode())
        if percent_float < 0 or percent_float > 100:
            logging.info(f'Skipping "{progress_report}" as the percent is not in [0, 100]')
            return False

        percent_int = int(round(percent_float))
        logging.debug(f'Updating progress to {percent_int} percent [tag={self.location_tag}]')

        self.progress = {'progress-message': message_data,
                         'progress-percent': percent_int,
                         'progress-sequence': self.sequence_counter.increment_and_get()}
        return True

    def retrieve_progress_states(self):
        """Generates the progress states by tailing the target_file.
        It tails a target file (using the tail() method) and uses the provided
        regex to find a match for a progress message. The regex is expected to
        generate two components in the match: the progress percent as an int and
        a progress message string. When such a message is found, this method
        yields the current progress as a dictionary.

        Note: This function must rethrow any OSError exceptions that it encounters.

        Returns
        -------
        An incrementally generated list of progress states.
        """
        if self.progress_regex_string:
            sleep_time_ms = 50
            for line in self.tail(sleep_time_ms):
                try:
                    progress_report = self.match_progress_update(line)
                    if progress_report is not None:
                        if self.__update_progress(progress_report):
                            yield self.progress
                except Exception:
                    logging.exception(f'Skipping progress line due to error: {line}')


class ProgressTracker(object):
    """Helper class to track progress messages from the specified location."""

    def __init__(self, config, counter, progress_updater, location, location_tag):
        """Launches the threads that track progress and send progress updates to the driver.

        Parameters
        ----------
        config: cook.sidecar.config.ExecutorConfig
            The current executor config.
        progress_updater: ProgressUpdater
            The progress updater used to send the progress messages
        counter: ProgressSequenceCounter
            The sequence counter
        location: string
            The target location to read for progress messages
        location_tag: string
            A tag to identify the target location."""
        self.location_tag = location_tag
        self.stop_event = Event()
        self.watcher = ProgressWatcher(location, location_tag, counter, config.max_bytes_read_per_line,
                                       config.progress_regex_string, self.stop_event)
        self.updater = progress_updater
        self.tracker_thread = Thread(target=self.track_progress, args=(), daemon=True)

    def start(self):
        """Launches a thread that starts monitoring the progress location for progress messages."""
        logging.info(f'Starting progress monitoring from [tag={self.location_tag}]')
        self.tracker_thread.start()

    def stop(self):
        """Signal this progress tracker thread to stop."""
        logging.info(f'Stop signal received on progress monitoring thread [tag={self.location_tag}]')
        self.stop_event.set()

    def wait(self):
        """Wait for this progress tracker to complete."""
        while self.tracker_thread.is_alive():
            self.tracker_thread.join(0.01)
        self.force_send_progress_update()
        logging.info(f'Completed progress monitoring from [tag={self.location_tag}]')

    def track_progress(self):
        """Retrieves and sends progress updates using send_progress_update_fn."""
        try:
            for current_progress in self.watcher.retrieve_progress_states():
                self.updater.send_progress_update(current_progress)
        except Exception:
            logging.exception(f'Exception while tracking progress [tag={self.location_tag}]')

    def force_send_progress_update(self):
        """Retrieves the latest progress message and attempts to force send it to the scheduler."""
        latest_progress = self.watcher.current_progress()
        self.updater.send_progress_update(latest_progress, force_send=True)



#!/usr/bin/env python3

from setuptools import setup

from cook.sidecar import version

requirements = [
    'flask~=1.1.0',
    'gunicorn~=20.1.0',
    'requests~=2.26.0',
]

test_requirements = [
]

setup(
    name='cook_sidecar',
    version=version.VERSION,
    description="Two Sigma's Cook Sidecar",
    long_description="The Cook Sidecar provides sandbox file access and progress reporting.",
    packages=['cook.sidecar'],
    entry_points={'console_scripts': ['cook-sidecar = cook.sidecar.__main__:main']},
    install_requires=requirements,
    tests_require=test_requirements
)



#
#  Copyright (c) 2020 Two Sigma Open Source, LLC
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#  IN THE SOFTWARE.
#

import logging
import os
import sys

def init_logging():
    log_level = os.environ.get('EXECUTOR_LOG_LEVEL', 'INFO')
    logging.basicConfig(level = log_level,
                        stream = sys.stderr,
                        format='%(asctime)s %(levelname)s %(message)s')
